import json
import os
from pathlib import Path
from typing import Any, Dict, List

import requests

from .exceptions import NoArgsException
from .utils import Site, TechnologiesProcessor, read_json

ABS_PATH = Path(__file__).resolve().parent


class Pywappalyzer:
    def __init__(self):
        self.technologies = read_json(
            str(os.path.join(ABS_PATH, "db/technologies.json"))
        )
        self.categories = read_json(str(os.path.join(ABS_PATH, "db/categories.json")))
        self.technologies_link = "https://raw.githubusercontent.com/AliasIO/wappalyzer/master/src/technologies.json"

    def analyze(self, url: str) -> Dict[str, List[str]]:
        """
        Analyze technologies of web page
        :param url: Url of web site
        :return: Dictionary of technologies.
            {"Web servers": ["Nginx"], ...}
        """
        site = Site(url=url)
        if site.headers is None:
            raise NoArgsException("No site headers")

        headers: Dict[str, Any] = site.headers
        processor = TechnologiesProcessor(
            headers=headers,
            site=site,
            technologies=self.technologies,
            categories=self.categories,
        )

        return processor.analyze()

    def use_latest(self) -> None:
        """
        Update technologies/categories json
        :return:
        """
        response = requests.get(self.technologies_link)
        items = response.json()

        with open(os.path.join(ABS_PATH, "db/categories.json"), "w+") as f:
            f.truncate()
            json.dump(items["categories"], f, ensure_ascii=False, indent=4)

        with open(os.path.join(ABS_PATH, "db/technologies.json"), "w+") as f:
            f.truncate()
            json.dump(items["technologies"], f, ensure_ascii=False, indent=4)



import json

import pytest


class Helpers:
    def read_json(self, filename: str) -> dict:
        with open(filename, "r") as f:
            data = json.load(f)
        return data

    def cookie_msg(self):
        return (
            "laravel_session=eyJpdiI6InB1UHF1N1pEUFU3R0xxZU1rNzVMVkE9PSIsIn"
            "ZhbHVlIjoiTVBZXC9aSHkreWhUbEhIM1E2NnlUVG0xSFdER2MzRDk0RWtcLzM2"
            "SHdCN05ZRUFQSmd1XC95XC9DalVPelJYRTQra2ciLCJtYWMiOiI5OTQ4MGYyNz"
            "cxYzc4M2RjYWQ0NTg5Y2IzZWY5ODg1NWU3ZjdlMjIyNGNlOTI3ZjQzMDNmOGQz"
            "ZjQ2ODQxY2ZlIn0%3D; expires=Fri, 01-Jan-2021 12:31:22 GMT; "
            "Max-Age=7200; path=/; httponly"
        )

    def pattern_msg(self):
        return r"^(.+)?\;version:\1"

    def headers_msg(self):
        return {
            "Set-Cookie": self.cookie_msg(),
            "Connection": "keep-alive",
            "Content-Length": "50426",
            "Server": "nginx",
            "Content-Type": "text/html; charset=utf-8",
            "X-Frame-Options": "DENY",
            "Via": "1.1 vegur, 1.1 varnish, 1.1 varnish",
            "Accept-Ranges": "bytes",
            "Date": "Sat, 02 Jan 2021 13:00:54 GMT",
            "Age": "1610",
            "X-Served-By": "cache-bwi5146-BWI, cache-hhn4038-HHN",
            "X-Cache": "HIT, HIT",
            "X-Cache-Hits": "1, 7",
            "X-Timer": "S1609592455.808592,VS0,VE0",
            "Vary": "Cookie",
            "Strict-Transport-Security": "max-age=63072000; includeSubDomains",
        }

    def read_txt(self, filename: str, *, as_bytes: bool = False):
        with open(filename, "r") as f:
            if as_bytes:
                return f.read().encode()
            return f.read()

    def analyze_headers_msg(self):
        return {
            "Web frameworks": ["Laravel"],
            "Web servers": ["Nginx"],
            "Reverse proxies": ["Nginx"],
            "Caching": ["Varnish"],
        }

    def sort_dict_lists(self, value):
        result = {}
        for key, val in value.items():
            result[key] = sorted(val)

        return result

    def result_analyze_msg(self):
        return {
            "Web frameworks": ["Laravel"],
            "Web servers": ["Nginx"],
            "Reverse proxies": ["Nginx"],
            "Caching": ["Varnish"],
            "Analytics": ["Google Analytics"],
            "JavaScript libraries": ["jQuery", "jQuery UI", "Modernizr"],
        }


@pytest.fixture
def helpers():
    return Helpers()



from http.cookies import Morsel

import pytest
from bs4.element import Tag
from requests import Session

from pywappalyzer.utils import (
    Site,
    TechnologiesProcessor,
    parse_cookie,
    read_json,
    resolve_pattern,
)


@pytest.fixture
def site(mock_get_html, mock_handle_js):
    site = Site(url="http://python.org/")
    return site


@pytest.fixture
def technologies_processor(site, helpers):
    site.meta["headers"] = helpers.headers_msg()
    return TechnologiesProcessor(
        headers=site.headers,
        site=site,
        technologies=helpers.read_json("samples/example_technologies.json"),
        categories=helpers.read_json("samples/example_categories.json"),
    )


@pytest.fixture
def mock_requests(mocker, helpers):
    mock_requests = mocker.patch.object(Session, "get")
    mock_requests.return_value.headers = helpers.headers_msg()
    return mock_requests


@pytest.fixture
def mock_get_html(mocker, helpers):
    mock_get_html = mocker.patch.object(Site, "get_html")
    mock_get_html.return_value = helpers.read_txt(
        "samples/example_html.txt", as_bytes=True
    )
    return mock_get_html


@pytest.fixture
def mock_handle_js(mocker, helpers):
    mock_handle_js = mocker.patch.object(Site, "handle_js")
    mock_handle_js.return_value = None
    return mock_handle_js


def test_parse_cookie(helpers):
    parsed_cookie = parse_cookie(helpers.cookie_msg())
    assert isinstance(parsed_cookie["laravel_session"], Morsel)


@pytest.mark.parametrize("filename", ["samples/example.json"])
def test_read_json(helpers, filename):
    data = read_json(filename)
    expected = helpers.read_json(filename)

    assert data == expected


def test_resolve_pattern(helpers):
    pattern = resolve_pattern(helpers.pattern_msg())
    assert pattern == r"^(.+)?"

    pattern = resolve_pattern(r"^(.+)?")
    assert pattern == r"^(.+)?"


def test_site_init(site, helpers):
    assert site.html == helpers.read_txt("samples/example_html.txt", as_bytes=True)
    assert site.js == {}


def test_get_headers(site, mock_requests, helpers):
    assert site.get_headers() == helpers.headers_msg()


def test_get_styles(site):
    assert all(
        True if isinstance(element, Tag) else False for element in site.get_styles()
    )


def test_get_host(site):
    assert site.get_host(url="https://google.com") == "google.com"


def test_get_scripts(site):
    assert all(
        True if isinstance(element, Tag) else False for element in site.get_scripts()
    )


def test_technologies_processor_analyze_headers(technologies_processor, helpers):
    assert technologies_processor.analyze_headers() == helpers.analyze_headers_msg()


def test_technologies_processor_analyze_styles(technologies_processor, helpers):
    assert technologies_processor.analyze_styles() == {}


def test_technologies_processor_analyze_scripts(technologies_processor, helpers):
    result = technologies_processor.analyze_scripts()
    assert result == {
        "Analytics": ["Google Analytics", "Google Analytics"],
        "JavaScript libraries": [
            "Modernizr",
            "jQuery",
            "jQuery",
            "jQuery UI",
            "jQuery UI",
            "jQuery UI",
            "jQuery UI",
        ],
    }


def test_technologies_processor_analyze(technologies_processor, helpers):
    result = technologies_processor.analyze()
    assert helpers.sort_dict_lists(result) == helpers.sort_dict_lists(
        helpers.result_analyze_msg()
    )



from dataclasses import dataclass
from unittest.mock import Mock

import pytest

from pywappalyzer.wappalyzer import Pywappalyzer


@dataclass
class Site:
    headers: dict


@pytest.fixture
def pywappalyzer():
    return Pywappalyzer()


@pytest.fixture
def mock_site(mocker):
    mock_site = mocker.patch("pywappalyzer.wappalyzer.Site")
    mock_site.return_value = Site(headers={})
    return mock_site


@pytest.fixture
def mock_open(mocker):
    mock_open = mocker.patch("builtins.open")
    mock_open.return_value.truncate.return_value = None
    return mock_open


@pytest.fixture
def mock_json(mocker):
    mock_json = mocker.patch("pywappalyzer.wappalyzer.json")
    mock_json.dump.return_value = None
    return mock_json


@pytest.fixture
def mock_requests_get():
    mock_requests_get = Mock()
    mock_requests_get.json.return_value = {"categories": {}, "technologies": {}}
    return mock_requests_get


@pytest.fixture
def mock_requests(mocker, mock_requests_get):
    mock_requests = mocker.patch("pywappalyzer.wappalyzer.requests")
    mock_requests.get.return_value = mock_requests_get
    return mock_requests_get


@pytest.fixture
def mock_technologies_processor(mocker, helpers):
    mock_technologies_processor = mocker.patch(
        "pywappalyzer.wappalyzer.TechnologiesProcessor"
    )
    mock_technologies_processor.return_value.analyze.return_value = (
        helpers.result_analyze_msg()
    )
    return mock_technologies_processor


def test_analyze(pywappalyzer, mock_site, mock_technologies_processor):
    result = pywappalyzer.analyze(url="https://python.org/")
    assert result == {
        "Web frameworks": ["Laravel"],
        "Web servers": ["Nginx"],
        "Reverse proxies": ["Nginx"],
        "Caching": ["Varnish"],
        "Analytics": ["Google Analytics"],
        "JavaScript libraries": ["jQuery", "jQuery UI", "Modernizr"],
    }


def test_use_latest(pywappalyzer, mock_json, mock_open, mock_requests):
    pywappalyzer.use_latest()



from platform import platform

import geckodriver_autoinstaller

from .utils import Site, TechnologiesProcessor  # noqa: F841,F401
from .wappalyzer import Pywappalyzer  # noqa: F841,F401

if "windows" in platform().lower():
    geckodriver_autoinstaller.install()



import json
import os
import re
from http.cookies import SimpleCookie
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup as bs  # noqa
from bs4 import ResultSet
from selenium import webdriver
from selenium.common.exceptions import JavascriptException
from selenium.webdriver.firefox.options import Options

ABS_PATH = Path(__file__).resolve().parent


def parse_cookie(cookie: str) -> dict:
    """
    Parse cookie string to dict
    :param cookie: Cookie string
    :return: Cookie's dictionary
    """
    _cookie: SimpleCookie = SimpleCookie()
    _cookie.load(cookie)

    cookies = {}
    for key, value in _cookie.items():
        cookies[key] = value

    return cookies


def read_json(path: str) -> dict:
    """
    Read json file
    :param path: File path
    :return: Dictionary from file
    """
    with open(path, "r") as f:
        data = json.load(f)
    return data


def resolve_pattern(pattern: str) -> str:
    """
    Parse regex pattern
    :param pattern: Regex pattern
    :return: Correct pattern
    """
    if ";version:" in pattern:
        _pattern = pattern.split(";version:")[0]
        return _pattern[:-1]
    return pattern


class Site:
    """
    Site interface for easy interacting with HTML/Headers etc.
    """

    def __init__(self, url: str, headers: Optional[Dict[str, Any]] = None) -> None:
        self.headers = headers
        if headers is None:
            self.headers = {
                "accept": "*/*",
                "user-agent": (
                    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
                    "(KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36"
                ),
            }

        self.url = url
        self.js: Dict[str, Dict[str, bool]] = {}
        self.html = self.get_html()
        self.meta = {
            "html": self.html,
            "js": self.js,
            "headers": self.get_headers(),
            "host": self.get_host(),
            "styles": self.get_styles(),
            "scripts": self.get_scripts(),
        }

    def get_html(
        self, url: Optional[str] = None, *, as_text: bool = False
    ) -> Union[bytes, str]:  # pragma: no cover
        """
        Scrape site's html
        :param url: Site's url
        :param as_text: Return html as string
        :return: Site's HTML as bytes or string
        """
        if url is None:
            url = self.url

        options = Options()
        options.add_argument("--headless")
        with webdriver.Firefox(options=options) as driver:
            driver.get(url)
            page_source = driver.page_source
            self.handle_js(driver)

        if as_text:
            return page_source
        return page_source.encode("utf-8")

    def handle_js(
        self, driver: Union[webdriver.Firefox, webdriver.Chrome]
    ) -> None:  # pragma: no cover
        """
        Handle JavaScript libraries which uses the web page
        :param driver: Selenium webdriver
        """
        technologies = read_json(str(os.path.join(ABS_PATH, "db/technologies.json")))

        for technology_key in technologies.keys():
            technology = technologies[technology_key]
            js = technology.get("js")

            if js is None:
                continue

            outputs = []
            temp_: Dict[str, Dict[str, bool]] = {}
            for script, pattern in js.items():
                try:
                    output = driver.execute_script(f"return {script}")
                    if output is None:
                        output = False
                    else:
                        output = True
                    outputs.append(output)

                except JavascriptException:
                    continue

                if temp_.get(technology_key) is None:
                    temp_[technology_key] = {}

                temp_[technology_key][script] = output

            if all(outputs):
                self.js = {**self.js, **temp_}

    def get_headers(self):
        """
        Get headers from response
        :return: Response headers
        """
        with requests.Session() as session:
            response = session.get(self.url, headers=self.headers)

        return response.headers

    def _get_elements(
        self, tag: str, attrs: Optional[Dict[str, str]] = None
    ) -> ResultSet:
        """
        Parse tags from site's html
        :param tag: Needed tag
        :param attrs: Tag attributes - {"class": "SomeClass"}
        :return: List of tag elements
        """
        if attrs is None:
            attrs = {}

        soup = bs(self.html, "lxml")
        elements = soup.find_all(tag, attrs=attrs)
        return elements

    def get_styles(self) -> list:
        """
        Get tags with "link" naming
        :return: List of tag elements
        """
        result = self._get_elements(tag="link") + self._get_elements(tag="style")
        return result

    def get_host(self, url: Optional[str] = None) -> str:
        """
        Parse host of url
        :param url: Url of site
        :return: Domain from URI
        """
        if url is None:
            url = self.url

        return urlparse(url).netloc

    def get_scripts(self) -> ResultSet:
        """
        Get tags with "script" naming
        :return: List of tag elements
        """
        return self._get_elements(tag="script")


class TechnologiesProcessor:
    """
    Handle technologies on web page
    """

    def __init__(
        self,
        headers: Dict[str, Any],
        site: Site,
        technologies: Dict[str, Any],
        categories: Dict[str, Any],
    ) -> None:
        self.headers = headers
        self.technologies = technologies
        self.site = site
        self.categories = categories
        self.technologies_keys = self.technologies.keys()

    def analyze_headers(self) -> Dict[str, List[str]]:
        """
        Analyze response headers
        :return: Dictionary with matched technologies
        """
        result: Dict[str, List[str]] = {}
        headers: dict = dict(self.site.meta["headers"])  # type: ignore
        headers_json: str = json.dumps(headers)

        for technology_key in self.technologies_keys:
            technology = self.technologies[technology_key]
            cookie = technology.get("cookies")
            meta = technology.get("headers")

            if cookie is not None and headers.get("Set-Cookie") is not None:
                cookie_keys = cookie.keys()
                response_cookie = parse_cookie(headers["Set-Cookie"])

                for cookie_key in cookie_keys:
                    if cookie_key in response_cookie:
                        for cat in technology["cats"]:
                            if result.get(self.categories[str(cat)]["name"]) is None:
                                result[self.categories[str(cat)]["name"]] = []

                            result[self.categories[str(cat)]["name"]].append(
                                technology_key
                            )

            if meta is None:
                continue

            for meta_key in meta.keys():
                if headers.get(meta_key) is None:
                    continue

                pattern = resolve_pattern(pattern=meta[meta_key])
                match = re.findall(pattern, headers_json, re.I)

                if (
                    len([el for el in match if len(el) > 0]) > 0
                    or technology_key.lower() == headers[meta_key].lower()
                    or technology_key.lower() in headers[meta_key].lower()
                ):
                    for cat in technology["cats"]:
                        if result.get(self.categories[str(cat)]["name"]) is None:
                            result[self.categories[str(cat)]["name"]] = []

                        result[self.categories[str(cat)]["name"]].append(technology_key)

        return result

    def analyze_styles(self) -> Dict[str, List[str]]:
        """
        Analyze web page styles
        :return: Dictionary with matched technologies
        """
        result: Dict[str, List[str]] = {}
        styles = self.site.meta["styles"]

        for technology_key in self.technologies_keys:
            technology = self.technologies[technology_key]
            patterns = (
                technology.get("html")
                if isinstance(technology.get("html"), list)
                else [technology.get("html", "")]
            )

            for style in styles:
                for pattern in patterns:
                    _pattern = resolve_pattern(pattern=pattern)
                    try:
                        match = re.findall(_pattern, " ".join(str(style).split()), re.I)
                    except re.error:
                        continue

                    if len([el for el in match if len(el) > 0]) > 0:
                        for cat in technology["cats"]:
                            if result.get(self.categories[str(cat)]["name"]) is None:
                                result[self.categories[str(cat)]["name"]] = []

                            result[self.categories[str(cat)]["name"]].append(
                                technology_key
                            )

        return result

    def analyze_scripts(self) -> Dict[str, List[str]]:
        """
        Analyze web page scripts
        :return: Dictionary with matched technologies
        """
        result: Dict[str, List[str]] = {}
        scripts = self.site.meta["scripts"]

        for technology_key in self.technologies_keys:
            technology = self.technologies[technology_key]
            patterns = (
                technology.get("scripts")
                if isinstance(technology.get("scripts"), list)
                else [technology.get("scripts", "")]
            )

            for script in scripts:
                for pattern in patterns:
                    _pattern = resolve_pattern(pattern=pattern)
                    try:
                        match = re.findall(_pattern, str(script), re.I)
                    except re.error:
                        continue

                    if self.site.meta["js"].get(technology_key) is not None:  # type: ignore
                        for cat in technology["cats"]:
                            if result.get(self.categories[str(cat)]["name"]) is None:
                                result[self.categories[str(cat)]["name"]] = []

                            result[self.categories[str(cat)]["name"]].append(
                                technology_key
                            )

                    if len([el for el in match if len(el) > 0]) > 0:
                        for cat in technology["cats"]:
                            if result.get(self.categories[str(cat)]["name"]) is None:
                                result[self.categories[str(cat)]["name"]] = []

                            result[self.categories[str(cat)]["name"]].append(
                                technology_key
                            )

        return result

    def analyze(self) -> Dict[str, List[str]]:
        """
        Analyze technologies on web page
        :return: Dictionary with matched technologies
        """
        styles = self.analyze_styles()
        headers = self.analyze_headers()
        scripts = self.analyze_scripts()

        techs = {
            key: list(set(value))
            for key, value in {**styles, **headers, **scripts}.items()
        }
        return techs
